#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "entities.h"

#include "math3d.h"

#include <cassert>

using namespace std;


//-----------------------------------


static
void setupMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (pos != NULL) glTranslatef(pos[0], pos[1], pos[2]);

  if (rot != NULL) {
    glRotatef(rot[0], 1, 0, 0);
    glRotatef(rot[1], 0, 1, 0);
    glRotatef(rot[2], 0, 0, 1);
  };

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


static
void setupInvMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (rot != NULL) {
    glRotatef(-rot[2], 0, 0, 1);
    glRotatef(-rot[1], 0, 1, 0);
    glRotatef(-rot[0], 1, 0, 0);
  };
  if (pos != NULL) glTranslatef(-pos[0], -pos[1], -pos[2]);

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


//--------------------------------


cGeospot::cGeospot(float* pos, float* rot, char* domain_name, unsigned int des)
{
  if (pos != NULL) memcpy(mPos.mV, pos, sizeof(float) * 3);
  if (rot != NULL) memcpy(mRot.mV, rot, sizeof(float) * 3);

	if (domain_name != NULL) mName = domain_name;
	mDesignation = des % 26;
}


	const char* cGeospot::translateDesignation(unsigned int des)
	{
		char* NATO[] =
		{
			"Alpha", "Bravo", "Charlie", "Delta", "Echo", "Foxtrot", "Golf", "Hotel", "India", "Juliett", "Kilo", "Lima", "M", "November", "O", "Papa",
			"Quibec", "Romeo", "Sierra", "Tango", "Uniform", "V", "W", "X", "Yankee", "Zulu"
		};
		if (des > 25) return "";
		return NATO[des];
	}


//--------------------------------


cWeapon::cWeapon()
{
  mCallGroup     = cWorld::GIGA;
  mNAmmo         = 0;
  mTAmmo         = NULL;
  mTime          = 0;
  mHeat          = 0;
  mMountfv       = NULL;
  mBasefv        = NULL;
}


void cWeapon::copySelf(cWeapon *source)
{
  ((cObject *)this)->copySelf(source);
  mBasefv        = source->mBasefv;
  mMountfv       = source->mMountfv;
  mNAmmo         = source->mNAmmo;
  mTAmmo         = source->mTAmmo;
  mTime          = source->mTime;
  mHeat          = source->mHeat;
}


void cWeapon::onSpawn()
{
	// Find Model-Object.
	cObjectFunction* objfun = new cObjectFunction_findName(NULL, "cParticleModel", 1);
  mWorld->callObjects(objfun, 0, true);
	assert(!objfun->collection.empty());
  cModel* mod = (cModel*) objfun->collection.front();
	objfun->collection.clear();
	delete objfun;

	// Register self for using that Model-Object.
  mModel = mod->addEntity(this, NULL);

	// Register a source matrix for particle spawning.
  cParticleModel* pm = (cParticleModel*) mModel->mModel;
  setupMatrix(mSource);
  pm->setSourceMatrix(mModel, mSource);
};


void cWeapon::fireSelf(cObject *target)
{

  cParticleModel* pm = (cParticleModel*) mModel->mModel;

	// Setup Barrel (source) Matrix.
  glPushMatrix();
  glLoadIdentity();
  glMultMatrixf(mBasefv);
  glMultMatrixf(mMountfv);
  glGetFloatv(GL_MODELVIEW_MATRIX, mSource);
  glPopMatrix();

	// Setup line for collision and collide with world.
  assert(mWorld != NULL);
  cMatrix m;
  memcpy(m.mM, mSource, sizeof(float) * 16);
  cVector a;
	a.Set(0, 0, 1);
  m.Apply(a);
  cVector b;
	b.Set(0, 0, 100);
  m.Apply(b);
	cObjectFunction* objfun = new cObjectFunction_lineCollision(a.mV, b.mV);
	mWorld->callObjects(objfun, 0, true);
	delete objfun;

  // Spawn the Flare-Ammo-Particle.
  pm->spawnParticle(mModel);

#if 0
  cVector v;

  cMatrix u;
  assert(mBasefv);
  memcpy(u.mM, mBasefv, sizeof(float) * 16);
  cMatrix w;
  assert(mMountfv);
  memcpy(w.mM, mMountfv, sizeof(float) * 16);

  cProjectile *shot = new cProjectile();

  if (mTAmmo) shot->copySelf(mTAmmo);
  shot->mTarget = target;
  shot->mSpeed = 150;
  shot->mPos.Set(0, 0, 1);
  shot->mDirection = shot->mPos;
  v.Set(0, 0, 1);
  shot->mDirection.Add(v);
  u.Apply(shot->mPos);
  w.Apply(shot->mPos);
  u.Apply(shot->mDirection);
  w.Apply(shot->mDirection);
  shot->mDirection.Sub(shot->mPos);
//  v = shot->mDirection;
//  v.Polar();
//  shot->mRot.Set(-mTower.mV[X], v.mV[Y], 0.0);
//  return;

  if (mWorld)
  {
    mWorld->spawnObject(shot);
//    this->transformSelf();
  };
#endif
}


//--------------------------------

/*
cProjectile::cProjectile(float* pos, float* rot)
{
  setupMatrix(mMatrix, rot, pos);

  mCallGroup     = cWorld::KILO;
  mGuided        = 0;
  mTarget        = NULL;
  mDestroyed     = false;
  mFlown         = 0;
}


void cProjectile::onSpawn()
{
	cObjectFunction* objfun = new cObjectFunction_findName(NULL, "cConstructModel", 1);
  mWorld->callObjects(objfun, 0, true);
	assert(!objfun->collection.empty());
  cConstructModel* mod = (cConstructModel*) objfun->collection.front();
	delete objfun;
  mModel = mod->addEntity(this, "data/objects/shot0.txt");
//  mod->setPlacementMatrix(mModel, this->mMatrix);
  mod->setRotator(mModel, this->mRot.mV);
  mod->setTranslator(mModel, this->mPos.mV);
}


void cProjectile::Fly(float spf)
{
  float a = mSpeed / 100.0 * spf * 16;
  float sp = a / 5.0;
#if 0
  const float e = 0.0174545;
  mPos.mV[X] += sp*sin(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
  mPos.mV[Y] -= sp*sin(mRot.mV[X]*e);
  mPos.mV[Z] += sp*cos(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
#endif
  cVector v = mDirection;
  v.Mul(a);
  mPos.Add(v);
//  if (mPos.mV[Y] < 0) mWorld->fragObject(this);
  mFlown += a;
//  if (mFlown > 100) mWorld->fragObject(this);
  setupMatrix(mMatrix, NULL, mPos.mV);
}


void cProjectile::animateSelf(float spf)
{
  if (mFragged) return;
  if (mController) mController->processSelf();
#if 0
  if (mWorld)
  {
    cObject *object;
    list<cObject*> objlist;
#if 1
    if (mWorld->scanSphere(mPos.mV, mParts[1]->mSphere.mR, object, (1L << 0L), true))
    {
      object->damageSphere(mParts[1]->mSphere, 0.001);
      mWorld->fragObject(this);
    };
#else
    if (mWorld->scanSphere(mPos.mV, mParts[1]->mSphere.mR, &objlist, (1L << 0L), true))
    {
      list<cObject*>::iterator it;
      for (it = objlist.begin(); it != objlist.end();)
      {
        object = *it; it++;
        if (object->mFragged) continue;
        if (object != this)
        {
          object->damageSphere(mParts[1]->mSphere, 0.001);
          mWorld->fragObject(this);
        };
      };
    };
    assert(0);
#endif
  };
#endif
  Fly(spf);
}


bool cProjectile::scanSphere(cSphere sphere)
{
  return false;
};
*/

//--------------------------------


cBuilding::cBuilding(float* pos, float* rot)
{
  if (pos != NULL) memcpy(mPos.mV, pos, sizeof(float) * 3);
  if (rot != NULL) memcpy(mRot.mV, rot, sizeof(float) * 3);
}


void cBuilding::onSpawn()
{
	// Find Model-Object.
	cObjectFunction* objfun = new cObjectFunction_findName(NULL, "cConstructModel", 1);
  mWorld->callObjects(objfun, 0, true);
	assert(!objfun->collection.empty());
  cConstructModel* mod = (cConstructModel*) objfun->collection.front();
	objfun->collection.clear();
	delete objfun;

	// Register self for using that Model-Object.
  mModel = mod->addEntity(this, "data/objects/build0.txt");

	// Register position and rotation vector for Model-Object.
  mod->setRotator(mModel, this->mRot.mV);
  mod->setTranslator(mModel, this->mPos.mV);
}


//--------------------------------


cTree::cTree(float* pos, float* rot, char* descriptor)
{
  if (pos != NULL) memcpy(mPos.mV, pos, sizeof(float) * 3);
  if (rot != NULL) memcpy(mRot.mV, rot, sizeof(float) * 3);

  if (descriptor != NULL) desc = descriptor;

  mGroupFlags |= (1L << 1L);
}


void cTree::onSpawn()
{
	// Find Model-Object.
	cObjectFunction* objfun = new cObjectFunction_findName(NULL, "cPlantModel", 1);
  mWorld->callObjects(objfun, 0, true);
	assert(!objfun->collection.empty());
  cPlantModel* mod = (cPlantModel*) objfun->collection.front();
	objfun->collection.clear();
	delete objfun;

	// Register self for using that Model-Object.
  char* descriptor = NULL;
  if (desc.size() > 0) descriptor = (char*) desc.c_str();
  mModel = mod->addEntity(this, descriptor);

	// Register position and rotation vector for Model-Object.
  mod->setRotator(mModel, this->mRot.mV);
  mod->setTranslator(mModel, this->mPos.mV);
}


//-----------------------------------


cMech::cMech(float* pos, float* rot)
{
  if (pos != NULL) memcpy(mPos.mV, pos, sizeof(float) * 3);
  if (rot != NULL) memcpy(mRot.mV, rot, sizeof(float) * 3);

  mGroupFlags |= (1L << 0L);

  mRotators = new float[MAX_MECH_PARTS][3];
  for (int i = 0; i < MAX_MECH_PARTS; i++)
  {
    memset(mRotators, 0, sizeof(float) * 3);
  };
}


void cMech::onSpawn()
{
	// Find Model-Object.
	cObjectFunction* objfun = new cObjectFunction_findName(NULL, "cConstructModel", 1);
  mWorld->callObjects(objfun, 0, true);
	assert(!objfun->collection.empty());
  cConstructModel* mod = (cConstructModel*) objfun->collection.front();
	delete objfun;

	// Register self for using that Model-Object.
  mModel = mod->addEntity(this, "data/objects/mech0.txt");

	// Register position and rotation vector for Model-Object.
  mod->setRotator(mModel, this->mRot.mV);
  mod->setTranslator(mModel, this->mPos.mV);

	// Register some more rotation vectors for each limb you want to use.
  mod->setConstructRotator(mModel, "CTorsor", mRotators[CTorsor]);
  mod->setConstructRotator(mModel, "LUpLeg", mRotators[LUpLeg]);
  mod->setConstructRotator(mModel, "LLoLeg", mRotators[LLoLeg]);
  mod->setConstructRotator(mModel, "RUpLeg", mRotators[RUpLeg]);
  mod->setConstructRotator(mModel, "RLoLeg", mRotators[RLoLeg]);
}


float* cMech::getPilotMatrix()
{
  glPushMatrix();
  glLoadIdentity();

  cConstructModel* co = (cConstructModel*) mModel->mModel;
  glMultMatrixf(co->getConstructInvMatrix(mModel, "Head"));
  glMultMatrixf(co->getInvMatrix(mModel));

  glGetFloatv(GL_MODELVIEW_MATRIX, mPilot);
  glPopMatrix();
  return mPilot;
}


void cMech::Pattern(float t, char *pattern)
{
  int l = strlen(pattern);
  int s = int(t/90.0) % l;
  char c = pattern[s];
  switch (c) {
    case 'l': ChasisLR(+5); break;
    case 'r': ChasisLR(-5); break;
    case 'n': break;
    default:  break;
  };
}


void cMech::ChasisLR(float r)
{
  const float limit = 1.75;
  float e = r;
  if (e>+limit) e = +limit;
  if (e<-limit) e = -limit;
  mRot.mV[Y] += e;
  mRot.mV[Y] = fmod(mRot.mV[Y], 360);
  if (mRot.mV[Y] < 0) mRot.mV[Y] += 360;
}


void cMech::ChasisUD(float r)
{
  float e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot.mV[X] += e;
  mRot.mV[X] = fmod(mRot.mV[X], 360);
  if (mRot.mV[X] < 0) mRot.mV[X] += 360;
}


void cMech::TowerLR(float r)
{
  float e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[Y] += e;
  if (mTwr.mV[Y] > +75) mTwr.mV[Y] = +75;
  if (mTwr.mV[Y] < -75) mTwr.mV[Y] = -75;
}


void cMech::TowerUD(float r)
{
  float e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr.mV[X] += e;
  if (mTwr.mV[X] > +75) mTwr.mV[X] = +75;
  if (mTwr.mV[X] < -75) mTwr.mV[X] = -75;
}


void cMech::Throttle(float percent)
{
  mSpeed = percent;
}


void cMech::Jet(float percent)
{
  mJetOn = percent;
}


void cMech::fireAllWeapons(cObject *target)
{
  list<cWeapon *>::iterator it;
  cWeapon *weapon;
  for (it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    weapon = *it;
    if (!weapon->getFragged()) weapon->fireSelf(target);
  };
}


void cMech::fireWeapon(int n, cObject *target)
{
  list<cWeapon *>::iterator it;
  cWeapon *weapon;
  int i = 0;
  for (it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    if (i == n)
    {
      weapon = *it;
      if (!weapon->getFragged()) weapon->fireSelf(target);
    };
    i++;
  };
}


void cMech::mountWeapon(char* point, cWeapon *weapon)
{
  cWeapon *wpn = new cWeapon;
  wpn->copySelf(weapon);
  cConstructModel* co = (cConstructModel*) mModel->mModel;
  wpn->mBasefv = co->getMatrix(mModel);
  wpn->mMountfv = co->getConstructMatrix(mModel, point);
  assert(wpn->mMountfv != NULL);

	// The copy of the weapon isn't spawned but it needs to know the world.
	// Hm, in later versions the weapon better just knows the mech instance.
  wpn->setWorld(mWorld);
  mWeapons.push_back(wpn);
  wpn->onSpawn();
}


void cMech::Walk(float spf)
{
  float a = mSpeed / 100.0;
  float sp = a * spf * 5.0;
  float jt = mJetOn / 100.0 * spf * 15.0;
  const float e = 0.0174545;
  mTimer += spf;
  float t = mTimer * 180;
  a = a*0.75;

  float vec[3] = { 0, 0, 0 };

  memcpy(mRotators[CTorsor], mTwr.mV, sizeof(float) * 3);

  if (mPos.mV[Y] < 0.5) {
    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (+sin(t*e)*30)*a;
    memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (-sin(t*e)*30)*a;
    memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (+sin(t*e*xsgn(a))*15+15)*fabs(a);
    memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (-sin(t*e*xsgn(a))*15+15)*fabs(a);
    memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
  } else {
    memset(vec, 0, sizeof(float) * 3);
    vec[X] = -10;
    memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = -10;
    memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = +20;
    memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = +20;
    memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
  };

  mPos.mV[X] += sp*sin(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);
  mPos.mV[Y] += jt;
  mPos.mV[Z] += sp*cos(mRot.mV[Y]*e)*cos(mRot.mV[X]*e);

  mPos.mV[X] -= mWorld->getGravity(X) * spf;
  mPos.mV[Y] -= mWorld->getGravity(Y) * spf;
  mPos.mV[Z] -= mWorld->getGravity(Z) * spf;

  // if ground contact.
  if (mPos.mV[Y] < 0.5) mPos.mV[Y] += sin((t*2.0 - 10)*e)*sp/2;
  if (mPos.mV[Y] < 0) mPos.mV[Y] = 0;
}


void cMech::animateSelf(float spf)
{
  if (mController) mController->processSelf();
  Walk(spf);
}


int cMech::handleMsg(int msg, long prma, long prmb, long prmc)
{
  if (msg == 0)
  {
    char *_cmd = (char *) prma;
    string cmd = _cmd;
    string ref;

    ref = "weaponcount";
    if (cmd == ref)
    {
      return mWeapons.size();
    };

    ref = "fire";
    if (cmd == ref)
    {
      fireWeapon((int) prmb, (cObject *) prmc);
      return 0;
    };

    ref = "fireall";
    if (cmd == ref)
    {
      fireAllWeapons((cObject *) prmb);
      return 0;
    };

    ref = "ChasisLR";
    if (cmd == ref) {
      ChasisLR((float) prmb);
      return 0;
    };

    ref = "ChasisUD";
    if (cmd == ref) {
      ChasisUD((float) prmb);
      return 0;
    };

    ref = "TowerLR";
    if (cmd == ref) {
      TowerLR((float) prmb);
      return 0;
    };

    ref = "TowerUD";
    if (cmd == ref) {
      TowerUD((float) prmb);
      return 0;
    };

    ref = "Throttle";
    if (cmd == ref) {
      Throttle((float) prmb);
      return 0;
    };

    ref = "Jet";
    if (cmd == ref) {
      Jet((float) prmb);
      return 0;
    };

    ref = "getTowerPtr";
    if (cmd == ref) {
      float **f = ((float **) prmb);
      *f = mTwr.mV;
      return 0;
    };
  };
  return ((cEntity *)this)->handleMsg(msg, prma, prmb, prmc);
}


//-----------------------------------

